@@ -8,12 +8,10 @@ See tunnel.h for some documentation about the programming interface.
 
 #include <time.h>
 #include <stdio.h>
+#include <netdb_.h>
 #include <fcntl.h>
-#include <netdb.h>
 #include <stdlib.h>
-#ifdef HAVE_SYS_POLL_H
-#include <sys/poll.h>
-#endif
+#include <sys/poll_.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 
@@ -22,6 +20,7 @@ See tunnel.h for some documentation about the programming interface.
 #include "common.h"
 
 /* #define IO_COUNT_HTTP_HEADER */
+/* #define USE_SHUTDOWN */
 
 #define READ_TRAIL_TIMEOUT (1 * 1000) /* milliseconds */
 #define ACCEPT_TIMEOUT 10 /* seconds */
@@ -87,16 +86,11 @@ struct tunnel
   size_t out_total_raw;
   size_t out_total_data;
   time_t out_connect_time;
+  int strict_content_length;
+  int keep_alive;
+  int max_connection_age;
 };
 
-#ifndef HAVE_ENDPROTOENT
-static void
-endprotoent (void)
-{
-  /* do nothing */
-}
-#endif /* HAVE_ENDPROTOENT */
-
 static const size_t sizeof_header = sizeof (Request) + sizeof (Length);
 
 static inline int
@@ -123,6 +117,7 @@ tunnel_is_client (Tunnel *tunnel)
   return !tunnel_is_server (tunnel);
 }
 
+#if 1
 static int
 get_proto_number (const char *name)
 {
@@ -138,6 +133,7 @@ get_proto_number (const char *name)
 
   return number;
 }
+#endif
 
 static int
 tunnel_in_setsockopts (int fd)
@@ -167,7 +163,7 @@ tunnel_in_setsockopts (int fd)
 		  &n);
       log_debug ("tunnel_out_setsockopts: SO_RCVLOWAT: %d", i);
     }
-#endif
+#endif /* SO_RCVLOWAT */
 
   return 0;
 }
@@ -176,31 +172,33 @@ static int
 tunnel_out_setsockopts (int fd)
 {
 #ifdef SO_SNDLOWAT
-  int tcp = get_proto_number ("tcp");
-
-  if (tcp != -1)
-    {
-      int i, n;
+  {
+    int tcp = get_proto_number ("tcp");
+    int i, n;
  
-      i = 1;
-      if (setsockopt (fd,
-		      tcp,
-		      SO_SNDLOWAT,
-		      (void *)&i,
-		      sizeof i) == -1)
-	{
-	  log_debug ("tunnel_out_setsockopts: non-fatal SO_SNDLOWAT error: %s",
-		     strerror (errno));
-	}
-      n = sizeof i;
-      getsockopt (fd,
-		  tcp,
-		  SO_SNDLOWAT,
-		  (void *)&i,
-		  &n);
-      log_debug ("tunnel_out_setsockopts: non-fatal SO_SNDLOWAT: %d", i);
-    }
-#endif
+    if (tcp != -1)
+      {
+	i = 1;
+	if (setsockopt (fd,
+			tcp,
+			SO_SNDLOWAT,
+			(void *)&i,
+			sizeof i) == -1)
+	  {
+	    log_debug ("tunnel_out_setsockopts: "
+		       "non-fatal SO_SNDLOWAT error: %s",
+		       strerror (errno));
+	  }
+	n = sizeof i;
+	getsockopt (fd,
+		    tcp,
+		    SO_SNDLOWAT,
+		    (void *)&i,
+		    &n);
+	log_debug ("tunnel_out_setsockopts: non-fatal SO_SNDLOWAT: %d", i);
+      }
+  }
+#endif /* SO_SNDLOWAT */
 
 #ifdef SO_LINGER
   {
@@ -227,8 +225,36 @@ tunnel_out_setsockopts (int fd)
     log_debug ("tunnel_out_setsockopts: SO_LINGER: onoff=%d linger=%d",
 	       l.l_onoff, l.l_linger);
   }
-#endif
+#endif /* SO_LINGER */
 
+#ifdef TCP_NODELAY
+  {
+    int tcp = get_proto_number ("tcp");
+    int i, n;
+ 
+    if (tcp != -1)
+      {
+	i = 1;
+	if (setsockopt (fd,
+			tcp,
+			TCP_NODELAY,
+			(void *)&i,
+			sizeof i) == -1)
+	  {
+	    log_debug ("tunnel_out_setsockopts: "
+		       "non-fatal TCP_NODELAY error: %s",
+		       strerror (errno));
+	  }
+	n = sizeof i;
+	getsockopt (fd,
+		    tcp,
+		    TCP_NODELAY,
+		    (void *)&i,
+		    &n);
+	log_debug ("tunnel_out_setsockopts: non-fatal TCP_NODELAY: %d", i);
+      }
+  }
+#else
 #ifdef SO_SNDBUF
   {
     int i, n;
@@ -251,7 +277,8 @@ tunnel_out_setsockopts (int fd)
 		&n);
     log_debug ("tunnel_out_setsockopts: SO_SNDBUF: %d", i);
   }
-#endif
+#endif /* SO_SNDBUF */
+#endif /* TCP_NODELAY */
 
 #ifdef SO_KEEPALIVE
   {
@@ -275,7 +302,7 @@ tunnel_out_setsockopts (int fd)
 		&n);
     log_debug ("tunnel_out_setsockopts: SO_KEEPALIVE: %d", i);
   }
-#endif
+#endif /* SO_KEEPALIVE */
 
   return 0;
 }
@@ -286,7 +313,7 @@ tunnel_out_disconnect (Tunnel *tunnel)
   if (tunnel_is_disconnected (tunnel))
     return;
 
-#ifdef DEBUG
+#ifdef DEBUG_MODE
   if (tunnel_is_client (tunnel) &&
       tunnel->bytes != tunnel->content_length + 1)
     log_error ("tunnel_out_disconnect: warning: "
@@ -340,7 +367,9 @@ tunnel_out_connect (Tunnel *tunnel)
 
   tunnel_out_setsockopts (tunnel->out_fd);
 
+#ifdef USE_SHUTDOWN
   shutdown (tunnel->out_fd, 0);
+#endif
 
   /* + 1 to allow for TUNNEL_DISCONNECT */
   n = http_post (tunnel->out_fd,
@@ -368,7 +397,7 @@ tunnel_out_connect (Tunnel *tunnel)
 static int
 tunnel_in_connect (Tunnel *tunnel)
 {
-  Http_response response;
+  Http_response *response;
   ssize_t n;
 
   log_verbose ("tunnel_in_connect()");
@@ -392,31 +421,42 @@ tunnel_in_connect (Tunnel *tunnel)
   if (http_get (tunnel->in_fd, &tunnel->dest) == -1)
     return -1;
 
+#ifdef USE_SHUTDOWN
   if (shutdown (tunnel->in_fd, 1) == -1)
     {
       log_error ("tunnel_in_connect: shutdown() error: %s",
 		 strerror (errno));
       return -1;
     }
+#endif
 
   n = http_parse_response (tunnel->in_fd, &response);
   if (n <= 0)
-    ;
-  else if (response.major_version != 1 ||
-	   (response.minor_version != 1 &&
-	    response.minor_version != 0))
+    {
+      if (n == 0)
+	log_error ("tunnel_in_connect: no response; peer "
+		   "closed connection");
+      else
+	log_error ("tunnel_in_connect: no response; error: %s",
+		   strerror (errno));
+    }
+  else if (response->major_version != 1 ||
+	   (response->minor_version != 1 &&
+	    response->minor_version != 0))
     {
       log_error ("tunnel_in_connect: unknown HTTP version: %d.%d",
-		 response.major_version, response.minor_version);
+		 response->major_version, response->minor_version);
       n = -1;
     }
-  else if (response.status_code != 200)
+  else if (response->status_code != 200)
     {
-      log_error ("tunnel_in_connect: HTTP error %d", response.status_code);
-      errno = http_error_to_errno (-response.status_code);
+      log_error ("tunnel_in_connect: HTTP error %d", response->status_code);
+      errno = http_error_to_errno (-response->status_code);
       n = -1;
     }
 
+  http_destroy_response (response);
+
   if (n > 0)
     {
 #ifdef IO_COUNT_HTTP_HEADER
@@ -427,7 +467,6 @@ tunnel_in_connect (Tunnel *tunnel)
     }
   else
     {
-      http_destroy_response (&response);
       return n;
     }
 
@@ -462,9 +501,46 @@ tunnel_write_request (Tunnel *tunnel, Request request,
     time (&t);
     if (tunnel_is_client (tunnel) &&
 	tunnel_is_connected (tunnel) &&
-	t - tunnel->out_connect_time > DEFAULT_CONNECTION_MAX_TIME)
+	t - tunnel->out_connect_time > tunnel->max_connection_age)
       {
 	char c = TUNNEL_DISCONNECT;
+
+	log_debug ("tunnel_write_request: connection > %d seconds old",
+		   tunnel->max_connection_age);
+
+	if (tunnel->strict_content_length)
+	  {
+	    int l = tunnel->content_length - tunnel->bytes - 1;
+
+	    log_debug ("tunnel_write_request: write padding (%d bytes)",
+		       tunnel->content_length - tunnel->bytes - 1);
+	    if (l > 3)
+	    {
+		char c;
+		short s;
+		int i;
+
+		c = TUNNEL_PADDING;
+  		tunnel_write_data (tunnel, &c, sizeof c);
+
+		s = htons(l-2);
+		tunnel_write_data (tunnel, &s, sizeof s);
+
+		l -= 2;
+		c = 0;
+	    	for (i=0; i<l; i++)
+  			tunnel_write_data (tunnel, &c, sizeof c);
+	    }
+	    else
+	    {
+		char c = TUNNEL_PAD1;
+		int i;
+		
+	    	for (i=0; i<l; i++)
+  			tunnel_write_data (tunnel, &c, sizeof c);
+	    }
+	  }
+
 	log_debug ("tunnel_write_request: closing old connection");
 	if (tunnel_write_data (tunnel, &c, sizeof c) <= 0)
 	  return -1;
@@ -524,7 +600,7 @@ tunnel_write_request (Tunnel *tunnel, Request request,
 			     sizeof network_length) == -1)
 	return -1;
 
-#ifdef DEBUG
+#ifdef DEBUG_MODE
       if (request == TUNNEL_DATA && debug_level >= 5)
 	{
 	  log_annoying ("tunnel_write_request: TUNNEL_DATA:");
@@ -556,7 +632,7 @@ tunnel_write_request (Tunnel *tunnel, Request request,
   log_annoying ("tunnel_write_data: out_total_raw = %u",
 		tunnel->out_total_raw);
 
-#ifdef DEBUG
+#ifdef DEBUG_MODE
   if (tunnel->bytes > tunnel->content_length)
     log_debug ("tunnel_write_request: tunnel->bytes > tunnel->content_length");
 #endif
@@ -593,7 +669,7 @@ tunnel_connect (Tunnel *tunnel)
 			    auth_data, sizeof auth_data) == -1)
     return -1;
 
-  if (tunnel_in_connect (tunnel) == -1)
+  if (tunnel_in_connect (tunnel) <= 0)
     return -1;
 
   return 0;
@@ -671,9 +747,12 @@ tunnel_close (Tunnel *tunnel)
   char buf[10240];
   ssize_t n;
 
-  log_debug ("tunnel_close: write padding (%d bytes)",
-	     tunnel->content_length - tunnel->bytes - 1);
-  tunnel_padding (tunnel, tunnel->content_length - tunnel->bytes - 1);
+  if (tunnel->strict_content_length)
+    {
+      log_debug ("tunnel_close: write padding (%d bytes)",
+		 tunnel->content_length - tunnel->bytes - 1);
+      tunnel_padding (tunnel, tunnel->content_length - tunnel->bytes - 1);
+    }
 
   log_debug ("tunnel_close: write TUNNEL_CLOSE request");
   tunnel_write_request (tunnel, TUNNEL_CLOSE, NULL, 0);
@@ -753,6 +832,7 @@ tunnel_read_request (Tunnel *tunnel, enum tunnel_request *request,
     }
   *request = req;
   tunnel->in_total_raw += n;
+  log_annoying ("request = 0x%x (%s)", req, REQ_TO_STRING (req));
 
   if (req & TUNNEL_SIMPLE)
     {
@@ -760,7 +840,7 @@ tunnel_read_request (Tunnel *tunnel, enum tunnel_request *request,
 		    tunnel->in_total_raw);
       log_debug ("tunnel_read_request:  %s", REQ_TO_STRING (req));
       *length = 0;
-      return 0;
+      return 1;
     }
 
   n = read_all (tunnel->in_fd, &len, 2);
@@ -768,22 +848,30 @@ tunnel_read_request (Tunnel *tunnel, enum tunnel_request *request,
     {
       log_error ("tunnel_read_request: error reading request length: %s",
 		 strerror (errno));
+      if (n == 0)
+	errno = EIO;
       return -1;
     }
   len = ntohs (len);
   *length = len;
   tunnel->in_total_raw += n;
+  log_annoying ("length = %d", len);
 
-  n = read_all (tunnel->in_fd, buf, (size_t)len);
-  if (n <= 0)
+  if (len > 0)
     {
-      log_error ("tunnel_read_request: error reading request data: %s",
-		 strerror (errno));
-      return -1;
+      n = read_all (tunnel->in_fd, buf, (size_t)len);
+      if (n <= 0)
+	{
+	  log_error ("tunnel_read_request: error reading request data: %s",
+		     strerror (errno));
+	  if (n == 0)
+	    errno = EIO;
+	  return -1;
+	}
+      tunnel->in_total_raw += n;
+      log_annoying ("tunnel_read_request: in_total_raw = %u",
+		    tunnel->in_total_raw);
     }
-  tunnel->in_total_raw += n;
-  log_annoying ("tunnel_read_request: in_total_raw = %u",
-		tunnel->in_total_raw);
 
   if (req == TUNNEL_DATA)
     log_verbose ("tunnel_read_request:  %s (%d)",
@@ -792,7 +880,7 @@ tunnel_read_request (Tunnel *tunnel, enum tunnel_request *request,
     log_debug ("tunnel_read_request:  %s (%d)",
 	       REQ_TO_STRING (req), len);
 
-  return 0;
+  return 1;
 }
 
 ssize_t
@@ -840,8 +928,11 @@ tunnel_read (Tunnel *tunnel, void *data, size_t length)
       return -1;
     }
 
-  if (tunnel_read_request (tunnel, &req, tunnel->buf, &len) == -1)
-    return -1;
+  if (tunnel_read_request (tunnel, &req, tunnel->buf, &len) <= 0)
+    {
+log_annoying ("tunnel_read_request returned <= 0, returning -1");
+      return -1;
+    }
 
   switch (req)
     {
@@ -995,7 +1086,7 @@ tunnel_accept (Tunnel *tunnel)
   while (tunnel->in_fd == -1 || tunnel->out_fd == -1)
     {
       struct sockaddr_in addr; 
-      Http_request request;
+      Http_request *request;
       struct pollfd p;
       ssize_t m;
       int len;
@@ -1025,22 +1116,18 @@ tunnel_accept (Tunnel *tunnel)
 	  return -1;
 	}
 
-      /* m = parse_header (s, &type); */
       m = http_parse_request (s, &request);
       if (m <= 0)
-	{
-	  http_destroy_request (&request);
-	  return m;
-	}
+	return m;
 
-      if (request.method == -1)
+      if (request->method == -1)
 	{
 	  log_error ("tunnel_accept: error parsing header: %s",
 		     strerror (errno));
 	  close (s);
 	}
-      else if (request.method == HTTP_POST ||
-	       request.method == HTTP_PUT)
+      else if (request->method == HTTP_POST ||
+	       request->method == HTTP_PUT)
 	{
 	  if (tunnel->in_fd == -1)
 	    {
@@ -1066,7 +1153,7 @@ tunnel_accept (Tunnel *tunnel)
 	      close (s);
 	    }
 	}
-      else if (request.method == HTTP_GET)
+      else if (request->method == HTTP_GET)
 	{
 	  if (tunnel->out_fd == -1)
 	    {
@@ -1126,7 +1213,7 @@ tunnel_accept (Tunnel *tunnel)
 	  close (s);
 	}
 
-      http_destroy_request (&request);
+      http_destroy_request (request);
     }
 
   if (tunnel->in_fd == -1 || tunnel->out_fd == -1)
@@ -1174,6 +1261,7 @@ tunnel_new_server (int port, size_t content_length)
   tunnel->in_total_data = 0;
   tunnel->out_total_raw = 0;
   tunnel->out_total_data = 0;
+  tunnel->strict_content_length = FALSE;
 
   tunnel->server_socket = server_socket (tunnel->dest.host_port, 1);
   if (tunnel->server_socket == -1)
@@ -1214,6 +1302,8 @@ tunnel_new_client (const char *host, int host_port,
   tunnel->dest.host_port = host_port;
   tunnel->dest.proxy_name = proxy;
   tunnel->dest.proxy_port = proxy_port;
+  tunnel->dest.proxy_authorization = NULL;
+  tunnel->dest.user_agent = NULL;
   /* -1 to allow for TUNNEL_DISCONNECT */
   tunnel->content_length = content_length - 1;
   tunnel->buf_ptr = tunnel->buf;
@@ -1222,6 +1312,7 @@ tunnel_new_client (const char *host, int host_port,
   tunnel->in_total_data = 0;
   tunnel->out_total_raw = 0;
   tunnel->out_total_data = 0;
+  tunnel->strict_content_length = FALSE;
 
   if (tunnel->dest.proxy_name == NULL)
     {
@@ -1255,3 +1346,84 @@ tunnel_destroy (Tunnel *tunnel)
 
   free (tunnel);
 }
+
+static int
+tunnel_opt (Tunnel *tunnel, const char *opt, void *data, int get_flag)
+{
+  if (strcmp (opt, "strict_content_length") == 0)
+    {
+      if (get_flag)
+	*(int *)data = tunnel->strict_content_length;
+      else
+	tunnel->strict_content_length = *(int *)data;
+    }
+  else if (strcmp (opt, "keep_alive") == 0)
+    {
+      if (get_flag)
+	*(int *)data = tunnel->keep_alive;
+      else
+	tunnel->keep_alive = *(int *)data;
+    }
+  else if (strcmp (opt, "max_connection_age") == 0)
+    {
+      if (get_flag)
+	*(int *)data = tunnel->max_connection_age;
+      else
+	tunnel->max_connection_age = *(int *)data;
+    }
+  else if (strcmp (opt, "proxy_authorization") == 0)
+    {
+      if (get_flag)
+	{
+	  if (tunnel->dest.proxy_authorization == NULL)
+	    *(char **)data = NULL;
+	  else
+	    *(char **)data = strdup (tunnel->dest.proxy_authorization);
+	}
+      else
+	{
+	  if (tunnel->dest.proxy_authorization != NULL)
+	    free ((char *)tunnel->dest.proxy_authorization);
+	  tunnel->dest.proxy_authorization = strdup ((char *)data);
+	  if (tunnel->dest.proxy_authorization == NULL)
+	    return -1;
+	}
+    }
+  else if (strcmp (opt, "user_agent") == 0)
+    {
+      if (get_flag)
+	{
+	  if (tunnel->dest.user_agent == NULL)
+	    *(char **)data = NULL;
+	  else
+	    *(char **)data = strdup (tunnel->dest.user_agent);
+	}
+      else
+	{
+	  if (tunnel->dest.user_agent != NULL)
+	    free ((char *)tunnel->dest.user_agent);
+	  tunnel->dest.user_agent = strdup ((char *)data);
+	  if (tunnel->dest.user_agent == NULL)
+	    return -1;
+	}
+    }
+  else
+    {
+      errno = EINVAL;
+      return -1;
+    }
+
+  return 0;
+}
+
+int
+tunnel_setopt (Tunnel *tunnel, const char *opt, void *data)
+{
+  return tunnel_opt (tunnel, opt, data, FALSE);
+}
+
+int
+tunnel_getopt (Tunnel *tunnel, const char *opt, void *data)
+{
+  return tunnel_opt (tunnel, opt, data, TRUE);
+}
